/*
*
* Test: Hashing with glib and with standard bib, tha means selfmade implementation.
*       At some points the selfmade impl. is kind of dirty.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <time.h>
#include <assert.h>

#define MAX_HASH 100000
#define MAX_STR 100
#define MULTI 33
#define TEST_FUNC 100000000                     /* Testvariable fr die Hashfunktionen*/
#define TEST_TABL 1000                          /* Testvariable fr die Einfge- und Suchoperation*/

struct names
{
    char vname[255];
    struct names *next;   
};

typedef struct {
    struct names *start;
    struct names *current;
    struct names *last;
} names_list;

names_list hash_table_sm[MAX_HASH];

/*
* Hashing, selfmade impl. 
* Daniel Bernstein Algo. hash << 5 = hash * 33
*/
int hashing(char *string)
{
    unsigned int hash;                          /* unsigned, da nur positive Zahlen */
    unsigned char *pointer;                     /* " */
    
    hash = 5381;                                /* nach dem Hashverfahren von */
    pointer = (unsigned char *) string;         /* pointer bekommt den Text bergebn */
    while(*pointer != '\0')                     /* '\0' Ende eines Strings */
    {
        hash =  MULTI * hash + *pointer;        /* Hashfunction */
        pointer++;                              /* der Pointer wird erhoeht,d.h auf den naechsten Buchstaben wird gezeigt */
    }
    return hash % MAX_HASH;                     /* % = Modulo */
}

/*
* Hashing, glib impl.
*/
int hashing_glib(unsigned char *string)
{
    unsigned int hash;
    hash = 0;
    hash = g_str_hash(string);
    return hash /*% MAX_HASH*/;
}

/*
* Insert a list of names in a Hash_table, which was impl. by glib
*/ 
int insertWithGlib()
{
    FILE *file;
    char data[100];
    char *name;
    
    file = fopen("Mnner Vornamen.txt","r");
    
    GHashTable *hash_table = g_hash_table_new(g_str_hash,g_str_equal);
    
    while(fgets(data,100,file) != NULL)
    {
        name = data;
        g_hash_table_insert(hash_table,name,name);   
    }
    
    g_hash_table_destroy(hash_table);
    fclose(file);
    
    return 0;   
}

/*
* Insert a list of names into a hashtable, which was impl. by myself
*/
int insertWithSelfmade()
{
    FILE *file;
    char data[100];
    char *name;
    unsigned int hash_ad;
   
    
    file = fopen("Mnner Vornamen.txt","r");
   
    while(fgets(data,100,file) != NULL)
    {
        name = data;
  
        hash_ad = hashing(name);
    
        names_list *list = &hash_table_sm[hash_ad];              //Zeiger auf Tabellenad.

        if(list->start == NULL)
        {
            list->start = malloc(sizeof(struct names));
            list->last = list->start;
        } else {
            list->last->next = malloc(sizeof(struct names));
            list->last = list->last->next;
        }
        if(!list->last) assert("Memmory");
     
        strcpy(list->last->vname, name);

        //printf("%s\n",hash_table_sm[hash_ad].start->vname);   //TEST
    }
    memset(hash_table_sm,0,sizeof(hash_table_sm));
    fclose(file);
    return 0;
}

/*
* Search-Impl. with glib
*/
float test_search_glib()
{
    FILE *file;
    char data[100];
    char *name;
    gchar *erg;
    int i;
    clock_t start, end;
    float total;
    
    file = fopen("Mnner Vornamen.txt","r");
    
    GHashTable *hash_table = g_hash_table_new(g_str_hash,g_str_equal);
    
    while(fgets(data,100,file) != NULL)
    {
        name = data;
        g_hash_table_insert(hash_table,name,name); 
    }
    
    fclose(file);
    
    
    start = clock();
    for( i = 0; i < TEST_TABL; i++)
    {
        file = fopen("Mnner Vornamen.txt","r");
        while(fgets(data,100,file) != NULL)
        {
            name = data;
            erg = g_hash_table_lookup(hash_table,name); 
        }
        fclose(file);
    }
    end = clock();
    total = ((float)(end - start))/CLOCKS_PER_SEC;
    g_hash_table_destroy(hash_table);
    return total;           
}

/*
* Search in the hashtable for a given string.
*/
struct names *search(char *string)
{
    names_list list;
    int hashad = hashing(string);
    
    list = hash_table_sm[hashad];
    
    list.current = list.start;
    
    while(list.current != NULL)
    {
        if(strcmp(string, list.current->vname)==0)
        {
            return list.current;
        }
        list.current = list.current->next;
    }
    
    return NULL;
}

float test_search_sm()
{
    FILE *file;
    char data[100];
    char *name;
    struct names *node;
    int i;
    clock_t start, end;
    float total;
    
    file = fopen("Mnner Vornamen.txt","r");
    // inserting names in the htable
    while(fgets(data,100,file) != NULL)
    {
        name = data;  
        int hash_ad = hashing(name);
    
        names_list *list = &hash_table_sm[hash_ad];              //Zeiger auf Tabellenad.

        if(list->start == NULL)
        {
            list->start = malloc(sizeof(struct names));
            list->last = list->start;
        } else {
            list->last->next = malloc(sizeof(struct names));
            list->last = list->last->next;
        }
        if(!list->last) assert("Memmory");
     
        strcpy(list->last->vname, name);
    }
    fclose(file);
    // end inserting
    
    // test search
    start = clock();
    for( i = 0; i < TEST_TABL; i++)
    {
        file = fopen("Mnner Vornamen.txt","r");
        while(fgets(data,100,file) != NULL)
        {
            name = data;
            node = search(name);
            //printf("%s\n",node->vname);   //TEST
        }
        fclose(file);   
    } 
    end = clock();
    total = ((float)(end - start))/CLOCKS_PER_SEC;
    return total;
}

/*
*   main-Function
*/
int main(int argc, char *argv[])
{
    int hash_norm;
    int hash_glib;
    int i;
    unsigned char *test_satz;
    clock_t start_n, end_n;
    clock_t start_g, end_g;
    float total_g, total_n;
    
    test_satz = "hallo";
    
    start_n = clock();
    for( i = 0; i < TEST_FUNC; i++)
    {     
        hash_norm = hashing(test_satz);
    }
    end_n = clock();
    start_g = clock();
    for( i = 0; i < TEST_FUNC; i++)
    {
        hash_glib = hashing_glib(test_satz);
    }
    end_g = clock();
    
    total_n = ((float)(end_n - start_n))/CLOCKS_PER_SEC;
    total_g = ((float)(end_g - start_g))/CLOCKS_PER_SEC;
    
    printf("Test: Glib vs Selfmade\n\n");
    printf("Variablen:\n\t MAX_HASH: %i\n\t TEST_FUNC(Iterationen): %i\n\t TEST_TABL(Iterationen): %i\n\n",MAX_HASH,TEST_FUNC,TEST_TABL);
    
    printf(" Zeitmessung Hashing, vom String: \"%s\": \n\n",test_satz);
    printf("\t Normal: %i, Time: %f \n",hash_norm,total_n);
    printf("\t Glib: %i, Time: %f \n",hash_glib % MAX_HASH,total_g);
    
    printf("\n\n Zeitmessung insert: \n");
    
    start_g = clock();
    for( i = 0; i < TEST_TABL; i++)
    {
        insertWithGlib();
    }
    end_g = clock();
    total_g = ((float)(end_g - start_g))/CLOCKS_PER_SEC;
    printf("\t Glib: %f\n",total_g); 


    start_n = clock();
    for( i = 0; i < TEST_TABL; i++)
    {
        insertWithSelfmade();
    }
    end_n = clock();
    total_n = ((double)(end_n - start_n))/CLOCKS_PER_SEC;
    printf("\t Normal: %f\n",total_n);


    printf("\n\n Zeitmessung search: \n");
    printf("\t Glib: %f\n",test_search_glib());
    printf("\t Normal: %f\n",test_search_sm());
    
    printf("\nDone...\n");
    getchar();
    
    return 0;
}

