#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#define MAX_HASH_OPEN 3902
#define MAX_HASH_CHAIN 100
#define MULTI 33
#define MAX_STRING 100
#define TEST 100

struct sth
{
    char sthName[255];
    struct sth *next;
};

typedef struct{
    struct sth *start;
    struct sth *current;
    struct sth *last;
} sth_list;

sth_list hash_table_chain[MAX_HASH_CHAIN];

char hash_table_open[MAX_HASH_OPEN][MAX_STRING];

/*
* Hashfunction h(k) = 33*k mod m
*/
int hashing(char *string,int m)
{
    unsigned int hashad;
    unsigned char *ptr;
    
    hashad = 0;
    ptr = (unsigned char *)string;
    while(*ptr != '\0')
    {
        hashad = MULTI * hashad + *ptr;  
        ptr++;  
    }   
    
    return hashad % m;
}

/*
* search in a open ad. HT
*/
char *search_open(char *string)
{
    int hashad = hashing(string,MAX_HASH_OPEN);
    while(hash_table_open[hashad][0] != 0)
    {
        if(strcmp(hash_table_open[hashad],string) == 0)
        {
            return hash_table_open[hashad];   
        }   
        else
        {
            hashad = (hashad + 1) % MAX_HASH_OPEN;   
        }
    } 
    return NULL;  
}

/*
* search in a table with linked lists
*/
struct sth *search_chain(char *string)
{
    sth_list list;
    int hashad = hashing(string,MAX_HASH_CHAIN);
    
    list = hash_table_chain[hashad];
    
    list.current = list.start;
    
    while(list.current != NULL)
    {
        if(strcmp(string, list.current->sthName)==0)
        {
            return list.current;
        }
        list.current = list.current->next;
    }
    
    return NULL;
}

/*
* insert an element into a open ad. Hashtable
*/
void insert_open(char *string)
{
    int hashad = hashing(string,MAX_HASH_OPEN);
        
    while(hash_table_open[hashad][0] != 0)
    {
        hashad = (hashad + 1) % MAX_HASH_OPEN;
    }
    
    strncpy(hash_table_open[hashad],string,MAX_STRING);
}

/*
* insert an element into a chained Hashtable
*/
void insert_chain(char *string)
{
    int hash_ad = hashing(string,MAX_HASH_CHAIN);
    
    sth_list *list = &hash_table_chain[hash_ad];              //Zeiger auf Tabellenad.

    if(list->start == NULL){
        list->start = malloc(sizeof(struct sth));
        list->last = list->start;
    } else {
        list->last->next = malloc(sizeof(struct sth));
        list->last = list->last->next;
    }
    if(!list->last) assert("Memmory");
     
    strcpy(list->last->sthName, string);
}

/*
* test for inserting an element into a HT with open ad. impl.
*/
float test_open()
{
    FILE *file;
    char *name;
    char data[100];
    clock_t start, end;
    float total;
    int i;
    
    start = clock();
    for( i = 0; i < TEST; i++)
    {
        file = fopen("Mnner Vornamen.txt","r");
        while(fgets(data,100,file) != NULL)
        {
            name = data;
            insert_open(name);
        }
        fclose(file);
        memset(hash_table_open,0,sizeof(hash_table_open));
    }
    end = clock();
    total = ((float)(end - start))/CLOCKS_PER_SEC;
    return total;   
}

/*
* test for searching an element in a HT with open ad. impl.
*/
float test_search_open()
{
    FILE *file;
    char *name;
    char data[100];
    clock_t start, end;
    float total;
    int i;
    
    file = fopen("Mnner Vornamen.txt","r");
    while(fgets(data,100,file) != NULL)
    {
        name = data;
        insert_open(name);
    }
    fclose(file);
    
    start = clock();
    for( i = 0; i < TEST; i++)
    {
        file = fopen("Mnner Vornamen.txt","r");
        while(fgets(data,100,file) != NULL)
        {
            name = data;
            search_open(name);
        }
        fclose(file);           
    }
    end = clock();
    total = ((float)(end - start))/CLOCKS_PER_SEC;
    memset(hash_table_open,0,sizeof(hash_table_open));
    return total; 
}

/*
* test for searching an element in HT with linked lists
*/
float test_search_chain()
{
    FILE *file;
    char *name;
    char data[100];
    clock_t start, end;
    float total;
    int i;
    
    file = fopen("Mnner Vornamen.txt","r");
    while(fgets(data,100,file) != NULL)
    {
        name = data;
        insert_chain(name);
    }
    fclose(file);
    
    start = clock();
    for( i = 0; i < TEST; i++)
    {
        file = fopen("Mnner Vornamen.txt","r");
        while(fgets(data,100,file) != NULL)
        {
            name = data;
            search_chain(name);
        }
        fclose(file);           
    }
    end = clock();
    total = ((float)(end - start))/CLOCKS_PER_SEC;
    memset(hash_table_open,0,sizeof(hash_table_open));
    return total; 
}
/*
* test for the Hashtable with the linked lists
*/
float test_chain()
{
    FILE *file;
    char *name;
    char data[100];
    clock_t start, end;
    float total;
    int i;
    
    start = clock();
    for( i = 0; i < TEST; i++)
    {
        file = fopen("Mnner Vornamen.txt","r");
        while(fgets(data,100,file) != NULL)
        {
            name = data;
            insert_chain(name);   
        } 
        fclose(file); 
        memset(hash_table_chain,0,sizeof(hash_table_chain));
    }
    end = clock();
    total = ((float)(end - start))/CLOCKS_PER_SEC;
    return total;
}

/*
* main-Function
*/
int main(int argc, char *argv[])
{
        
    printf("Test: Hashtable Open_Ad. vs Hashtable wirh linkedlists\n");
    printf("Values: Size HT_Op: %i, Size HT_LL: %i, Iterations: %i\n\n",MAX_HASH_OPEN,MAX_HASH_CHAIN,TEST);
  
    printf("Test \"insert\": \n");
    printf("\t HT_Op: %f\n",test_open());
    printf("\t HT_LL: %f\n",test_chain());
    printf("\n\"search\":\n");
    printf("\t HT_Op: %f\n",test_search_open());
    printf("\t HT_LL: %f\n",test_search_chain());
    printf("Done..\n");	
    return 0;
}

